@@ -5,9 +5,23 @@ UpSet.js Jupyter Widget
 """
 
 import typing as t
+from collections import OrderedDict
 
 from ipywidgets import DOMWidget, Layout, ValueWidget, register
-from traitlets import Bool, Dict, Enum, Float, Int, List, Tuple, Unicode, Union, default
+from ipywidgets.widgets.trait_types import Color
+from traitlets import (
+    Bool,
+    Dict,
+    Enum,
+    Float,
+    Int,
+    List,
+    Tuple,
+    Unicode,
+    Union,
+    default,
+    Instance,
+)
 
 from ._frontend import MODULE_NAME, MODULE_VERSION
 from ._model import (
@@ -21,23 +35,94 @@ from ._model import (
     UpSetSetUnion,
 )
 from ._generate import generate_unions, generate_intersections
+from ._array import compress_index_array
+
+
+__all__ = ["UpSetWidget"]
 
 
 def _sort_sets(
-    sets: t.Sequence[UpSetSetLike[T]], order_by: str, limit: t.Optional[int] = None,
-) -> t.List[UpSetSetLike[T]]:
-    out_list: t.List[UpSetSetLike[T]]
+    sets: t.Sequence[UpSetSet[T]], order_by: str, limit: t.Optional[int] = None,
+) -> t.List[UpSetSet[T]]:
+    key = None
     if order_by == "cardinality":
-        out_list = sorted(sets, key=lambda s: s.cardinality, reverse=True)
-    elif order_by == "degree":
-        out_list = sorted(sets, key=lambda s: s.degree, reverse=True)
+        key = lambda s: (-s.cardinality, s.name)
     else:
-        out_list = list(sets)
+        key = lambda s: s.name
+    out_list = sorted(sets, key=key)
     if limit is not None:
         return out_list[:limit]
     return out_list
 
 
+def _sort_combinations(
+    combinations: t.Sequence[UpSetSetCombination[T]],
+    sets: t.Sequence[UpSetSet[T]],
+    order_by: t.Union[str, t.Sequence[str]],
+    limit: t.Optional[int] = None,
+) -> t.List[UpSetSetCombination[T]]:
+    def to_key(order: str):
+        if order == "cardinality":
+            return lambda s: -s.cardinality
+        if order == "degree":
+            return lambda s: s.degree
+        if order == "group":
+            return lambda c: next(
+                (i for i, s in enumerate(sets) if s in c.sets), len(sets)
+            )
+        return lambda s: s.name
+
+    keys = (
+        [to_key(order_by)]
+        if isinstance(order_by, str)
+        else [to_key(v) for v in order_by]
+    )
+    out_list = sorted(combinations, key=lambda s: tuple([k(s) for k in keys]))
+    if limit is not None:
+        return out_list[:limit]
+    return out_list
+
+
+def _to_set_list(arr: t.List[UpSetSet], model: "UpSetWidget"):
+    return [
+        dict(
+            type=str(s.set_type),
+            name=s.name,
+            cardinality=s.cardinality,
+            elems=compress_index_array(model.elem_to_index[e] for e in s.elems),
+        )
+        for s in arr
+    ]
+
+
+def _to_combination_list(arr: t.List[UpSetSetCombination], model: "UpSetWidget"):
+    return [
+        dict(
+            type=str(s.set_type),
+            name=s.name,
+            cardinality=s.cardinality,
+            degree=s.degree,
+            set_names=[c.name for c in s.sets],
+            elems=compress_index_array(model.elem_to_index[e] for e in s.elems),
+        )
+        for s in arr
+    ]
+
+
+def _to_query_list(arr: t.List[UpSetQuery], model: "UpSetWidget"):
+    def _to_query(query: UpSetQuery):
+        query_dict: t.Dict = dict(name=query.name, color=query.color)
+        if query.set:
+            query_dict["set"] = dict(name=query.set.name, type=str(query.set.set_type))
+        else:
+            query_dict["elems"] = compress_index_array(
+                model.elem_to_index[e] for e in query.elems or []
+            )
+        return query_dict
+
+    return [_to_query(q) for q in arr]
+
+
 @register
 class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
     """UpSet Widget
@@ -50,8 +135,17 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
     _view_module = Unicode(MODULE_NAME).tag(sync=True)
     _view_module_version = Unicode(MODULE_VERSION).tag(sync=True)
 
-    mode: str = Enum(("hover", "click"), default_value="hover").tag(sync=True)
+    """
+    interactivity mode of the widget whether the plot is static, reacts on hover or click events
+    """
+    mode: str = Enum(("hover", "click", "static"), default_value="hover").tag(sync=True)
+    """
+    padding within the svg
+    """
     padding: float = Float(None, allow_none=True).tag(sync=True)
+    """
+    padding argument for scaleBand (0..1)
+    """
     bar_padding: float = Float(None, allow_none=True).tag(sync=True)
     dot_padding: float = Float(None, allow_none=True).tag(sync=True)
     width_ratios: t.Tuple[float, float, float] = Tuple(
@@ -62,29 +156,34 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
     ).tag(sync=True)
 
     elems: t.List[T] = List(default_value=[]).tag(sync=True)
-    _elem_to_index: t.Dict[T, int] = {}
+    elem_to_index: t.Dict[T, int] = {}
 
-    _sets_obj: t.List[UpSetSet[T]] = []
-    _sets: t.List[t.Mapping] = List(Dict(), default_value=[],).tag(sync=True)
+    attrs: t.Dict[str, t.List[float]] = Dict().tag(sync=True)
 
-    _combinations_obj: t.List[UpSetSetCombination[T]] = []
-    _combinations: t.List[t.Mapping] = List(Dict(), default_value=[],).tag(sync=True)
+    sets: t.List[UpSetSet[T]] = List(Instance(UpSetSet), default_value=[],).tag(
+        sync=True, to_json=_to_set_list
+    )
+
+    combinations: t.List[UpSetSetCombination[T]] = List(
+        Instance(UpSetSetCombination), default_value=[],
+    ).tag(sync=True, to_json=_to_combination_list)
 
     value: t.Union[None, t.Mapping, t.List[int]] = Union(
         (Dict(allow_none=True, default_value=None), List(Int(), default_value=[]))
     ).tag(sync=True)
     _selection: t.Union[None, t.FrozenSet[T], UpSetSetLike[T]] = None
 
-    _queries_obj: t.List[UpSetQuery[T]] = []
-    _queries: t.List[t.Mapping] = List(Dict(), default_value=[]).tag(sync=True)
+    queries: t.List[UpSetQuery[T]] = List(Instance(UpSetQuery), default_value=[]).tag(
+        sync=True, to_json=_to_query_list
+    )
 
     theme: str = Enum(("light", "dark"), default_value="light").tag(sync=True)
-    selection_color: str = Unicode(None, allow_none=True).tag(sync=True)
-    alternating_background_color: str = Unicode(None, allow_none=True).tag(sync=True)
-    color: str = Unicode(None, allow_none=True).tag(sync=True)
-    text_color: str = Unicode(None, allow_none=True).tag(sync=True)
-    hover_hint_color: str = Unicode(None, allow_none=True).tag(sync=True)
-    not_member_color: str = Unicode(None, allow_none=True).tag(sync=True)
+    selection_color: str = Color(None, allow_none=True).tag(sync=True)
+    alternating_background_color: str = Color(None, allow_none=True).tag(sync=True)
+    color: str = Color(None, allow_none=True).tag(sync=True)
+    text_color: str = Color(None, allow_none=True).tag(sync=True)
+    hover_hint_color: str = Color(None, allow_none=True).tag(sync=True)
+    not_member_color: str = Color(None, allow_none=True).tag(sync=True)
 
     bar_label_offset: float = Float(None, allow_none=True).tag(sync=True)
     set_name_axis_offset: float = Float(None, allow_none=True).tag(sync=True)
@@ -112,6 +211,51 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
         super().__init__(**kwargs)
         self.observe(self._sync_value, "value")
 
+    def copy(self) -> "UpSetWidget":
+        """
+        returns a copy of itself
+        """
+        clone = UpSetWidget[T]()  # pylint: disable=unsubscriptable-object
+
+        clone.mode = self.mode
+        clone.padding = self.padding
+        clone.bar_padding = self.bar_padding
+        clone.dot_padding = self.dot_padding
+        clone.width_ratios = self.width_ratios
+        clone.height_ratios = self.height_ratios
+        clone.elems = list(self.elems)
+        clone.elem_to_index = self.elem_to_index.copy()
+        clone.sets = list(self.sets)
+        clone.combinations = list(self.combinations)
+        clone.queries = list(self.queries)
+
+        clone.value = self.value
+        clone.selection = self.selection
+
+        clone.theme = self.theme
+        clone.selection_color = self.selection_color
+        clone.alternating_background_color = self.alternating_background_color
+        clone.color = self.color
+        clone.text_color = self.text_color
+        clone.hover_hint_color = self.hover_hint_color
+        clone.not_member_color = self.not_member_color
+
+        clone.bar_label_offset = self.bar_label_offset
+        clone.set_name_axis_offset = self.set_name_axis_offset
+        clone.combination_name_axis_offset = self.combination_name_axis_offset
+
+        clone.query_legend = self.query_legend
+        clone.export_buttons = self.export_buttons
+        clone.font_family = self.font_family
+        clone.font_sizes = self.font_sizes.copy()
+        clone.numeric_scale = self.numeric_scale
+        clone.band_scale = self.band_scale
+
+        clone.set_name = self.set_name
+        clone.combination_name = self.combination_name
+
+        return clone
+
     def _sync_value(self, evt: t.Any):
         self._selection = self._value_to_selection(evt["new"])
 
@@ -176,7 +320,7 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
         if value is None:
             self.value = None
         elif isinstance(value, (list, set, frozenset)):
-            self.value = [self._elem_to_index[e] for e in value]
+            self.value = [self.elem_to_index[e] for e in value]
         elif isinstance(value, UpSetSet):
             self.value = dict(type="set", name=value.name)
         else:
@@ -187,53 +331,17 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
         # unknown
         self.observe(self._sync_value, "value")
 
-    @property
-    def sets(self) -> t.Sequence[UpSetSet[T]]:
-        """
-        the UpSet list of sets
-        """
-        return self._sets_obj
-
-    @sets.setter
-    def sets(self, value: t.List[UpSetSet[T]]):
-        self._sets_obj = value
-        self._sets = [
-            dict(
-                type=str(s.set_type),
-                name=s.name,
-                cardinality=s.cardinality,
-                elems=[self._elem_to_index[e] for e in s.elems],
-            )
-            for s in self._sets_obj
-        ]
-
-    @property
-    def combinations(self) -> t.Sequence[UpSetSetCombination[T]]:
-        """
-        the UpSet list of set combinations
-        """
-        return self._combinations_obj
-
-    @combinations.setter
-    def combinations(self, value: t.List[UpSetSetCombination[T]]):
-        self._combinations_obj = value
-        self._combinations = [
-            dict(
-                type=str(s.set_type),
-                name=s.name,
-                cardinality=s.cardinality,
-                degree=s.degree,
-                set_names=[c.name for c in s.sets],
-                elems=[self._elem_to_index[e] for e in s.elems],
-            )
-            for s in self._combinations_obj
-        ]
-
     def on_selection_changed(self, callback):
         """
         add callback listener to listen for selection changes
         """
-        self.observe(lambda: callback(self.selection), "value")
+        self.observe(lambda _: callback(self.selection), "value")
+
+    def clear_queries(self):
+        """
+        deletes the list of queries
+        """
+        self.queries = []
 
     def append_query(
         self,
@@ -250,13 +358,7 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
             query = UpSetQuery[T](name, color, upset=upset)
         else:
             query = UpSetQuery[T](name, color, elems=elems or frozenset())
-        self._queries_obj.append(query)
-        query_dict: t.Dict = dict(name=query.name, color=query.color)
-        if query.set:
-            query_dict["set"] = dict(name=query.set.name, type=str(query.set.set_type))
-        else:
-            query_dict["elems"] = [self._elem_to_index[e] for e in query.elems or []]
-        self._queries.append(query_dict)
+        self.queries = self.queries + [query]
         return self
 
     @property
@@ -291,20 +393,6 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
         else:
             self.layout.height = value
 
-    @property
-    def queries(self):
-        """
-        current list of UpSet queries
-        """
-        return self._queries_obj
-
-    def clear_queries(self):
-        """
-        deletes the list of queries
-        """
-        self._queries = []
-        self._queries_obj = []
-
     @default("layout")
     def _default_layout(self):  # pylint: disable=no-self-use
         return Layout(height="400px", align_self="stretch")
@@ -322,17 +410,21 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
         for set_elems in sets.values():
             elems.update(set_elems)
         self.elems = sorted(elems)
-        self._elem_to_index = {e: i for i, e in enumerate(self.elems)}
+        self.elem_to_index = {e: i for i, e in enumerate(self.elems)}
 
         base_sets: t.List[UpSetSet[T]] = [
             UpSetSet[T](name=k, elems=frozenset(v)) for k, v in sets.items()
         ]
-        self.sets = t.cast(t.Any, _sort_sets(base_sets, order_by, limit))
+        self.clear_queries()
+        self.selection = None
+        self.sets = _sort_sets(base_sets, order_by, limit)
+        self.attrs = OrderedDict()
         return self.generate_intersections(order_by=order_by)
 
     def from_dataframe(
         self,
         data_frame: t.Any,
+        attributes: t.Union[t.Sequence[str], t.Any, None] = None,
         order_by: str = "cardinality",
         limit: t.Optional[int] = None,
     ):
@@ -340,14 +432,35 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
         generates the list of sets from a dataframe
         """
         self.elems = sorted(data_frame.index)
-        self._elem_to_index = {e: i for i, e in enumerate(self.elems)}
+        self.elem_to_index = {e: i for i, e in enumerate(self.elems)}
 
         def to_set(name: str, series):
             elems = series[series.astype(bool)].index
             return UpSetSet[T](name=name, elems=frozenset(elems))
 
-        base_sets = [to_set(name, series) for name, series in data_frame.items()]
-        self.sets = t.cast(t.Any, _sort_sets(base_sets, order_by, limit))
+        attribute_columns = attributes if isinstance(attributes, (list, tuple)) else []
+
+        base_sets = [
+            to_set(name, series)
+            for name, series in data_frame.items()
+            if name not in attribute_columns
+        ]
+        self.clear_queries()
+        self.selection = None
+        self.sets = _sort_sets(base_sets, order_by, limit)
+
+        if attributes is not None:
+            attribute_df = (
+                data_frame[attributes]
+                if isinstance(attributes, (list, tuple))
+                else attributes
+            )
+            self.attrs = OrderedDict(
+                [(name, series.tolist()) for name, series in attribute_df.items()]
+            )
+        else:
+            self.attrs = OrderedDict()
+
         return self.generate_intersections(order_by=order_by)
 
     def generate_intersections(
@@ -355,7 +468,7 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
         min_degree: int = 0,
         max_degree: t.Optional[int] = None,
         empty: bool = False,
-        order_by: str = "cardinality",
+        order_by: t.Union[str, t.Sequence[str]] = "cardinality",
         limit: t.Optional[int] = None,
     ):
         """
@@ -365,8 +478,8 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
             self.sets, min_degree, max_degree, empty, self.elems
         )
 
-        self.combinations = t.cast(
-            t.Any, _sort_sets(set_intersections, order_by, limit)
+        self.combinations = _sort_combinations(
+            set_intersections, self.sets, order_by, limit
         )
         return self
 
@@ -375,7 +488,7 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
         min_degree: int = 0,
         max_degree: t.Optional[int] = None,
         empty: bool = False,
-        order_by: str = "cardinality",
+        order_by: t.Union[str, t.Sequence[str]] = "cardinality",
         limit: t.Optional[int] = None,
     ):
         """
@@ -385,5 +498,5 @@ class UpSetWidget(ValueWidget, DOMWidget, t.Generic[T]):
             self.sets, min_degree, max_degree, empty, self.elems
         )
 
-        self.combinations = t.cast(t.Any, _sort_sets(set_unions, order_by, limit))
+        self.combinations = _sort_combinations(set_unions, self.sets, order_by, limit)
         return self
